@@ -49,16 +49,6 @@ static inline void nl_socket_free(struct nl_sock *h)
 	nl_handle_destroy(h);
 }
 
-static inline int __genl_ctrl_alloc_cache(struct nl_sock *h, struct nl_cache **cache)
-{
-	struct nl_cache *tmp = genl_ctrl_alloc_cache(h);
-	if (!tmp)
-		return -ENOMEM;
-	*cache = tmp;
-	return 0;
-}
-#define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
-
 #define NLE_MISSING_ATTR	14
 
 static inline void __nl_perror(int error, const char *s)
@@ -70,8 +60,7 @@ static inline void __nl_perror(int error, const char *s)
 
 struct nlnfc_state {
 	struct nl_sock *nl_sock;
-	struct nl_cache *nl_cache;
-	struct genl_family *nlnfc;
+	int nfc_id;
 	int mcid;
 };
 
@@ -189,20 +178,18 @@ int __near_netlink_get_adapters(void)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int err, family;
+	int err;
 
 	DBG("");
 
-	if (nfc_state == NULL || nfc_state->nlnfc == NULL)
+	if (nfc_state == NULL || nfc_state->nfc_id < 0)
 		return -ENODEV;
 
 	msg = nlmsg_alloc();
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			  NLM_F_DUMP, NFC_CMD_GET_DEVICE, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -217,21 +204,20 @@ out:
 	return err;
 }
 
-int __near_netlink_start_poll(int idx, uint32_t protocols)
+int __near_netlink_start_poll(int idx,
+				uint32_t im_protocols, uint32_t tm_protocols)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err = 0;
 
-	DBG("");
+	DBG("IM protos 0x%x TM protos 0x%x", im_protocols, tm_protocols);
 
 	msg = nlmsg_alloc();
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_START_POLL, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -239,7 +225,12 @@ int __near_netlink_start_poll(int idx, uint32_t protocols)
 	}
 
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
-	NLA_PUT_U32(msg, NFC_ATTR_PROTOCOLS, protocols);
+	if (im_protocols != 0) {
+		NLA_PUT_U32(msg, NFC_ATTR_IM_PROTOCOLS, im_protocols);
+		NLA_PUT_U32(msg, NFC_ATTR_PROTOCOLS, im_protocols);
+	}
+	if (tm_protocols != 0)
+		NLA_PUT_U32(msg, NFC_ATTR_TM_PROTOCOLS, tm_protocols);
 
 	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
 
@@ -254,7 +245,7 @@ int __near_netlink_stop_poll(int idx)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err = 0;
 
 	DBG("");
 
@@ -262,9 +253,7 @@ int __near_netlink_stop_poll(int idx)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_STOP_POLL, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -286,7 +275,7 @@ int __near_netlink_dep_link_up(uint32_t idx, uint32_t target_idx,
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err = 0;
 
 	DBG("");
 
@@ -294,9 +283,7 @@ int __near_netlink_dep_link_up(uint32_t idx, uint32_t target_idx,
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_DEP_LINK_UP, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -320,7 +307,7 @@ int __near_netlink_dep_link_down(uint32_t idx)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err = 0;
 
 	DBG("");
 
@@ -328,9 +315,7 @@ int __near_netlink_dep_link_down(uint32_t idx)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_DEP_LINK_DOWN, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -351,7 +336,7 @@ int __near_netlink_adapter_enable(int idx, near_bool_t enable)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err = 0;
 	uint8_t cmd;
 
 	DBG("");
@@ -360,14 +345,12 @@ int __near_netlink_adapter_enable(int idx, near_bool_t enable)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
 	if (enable == TRUE)
 		cmd = NFC_CMD_DEV_UP;
 	else
 		cmd = NFC_CMD_DEV_DOWN;
 
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, cmd, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -479,12 +462,12 @@ static int get_targets_handler(struct nl_msg *n, void *arg)
 	return 0;
 }
 
-static int nfc_netlink_event_targets(struct genlmsghdr *gnlh)
+static int nfc_netlink_event_targets_found(struct genlmsghdr *gnlh)
 {
 	struct nlattr *attr[NFC_ATTR_MAX + 1];
 	struct nl_msg *msg;
 	void *hdr;
-	int err, family;
+	int err;
 	uint32_t adapter_idx;
 
 	DBG("");
@@ -502,9 +485,7 @@ static int nfc_netlink_event_targets(struct genlmsghdr *gnlh)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_DUMP, NFC_CMD_GET_TARGET, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
@@ -522,6 +503,30 @@ nla_put_failure:
 	return err;
 }
 
+static int nfc_netlink_event_target_lost(struct genlmsghdr *gnlh)
+{
+	struct nlattr *attr[NFC_ATTR_MAX + 1];
+	uint32_t adapter_idx, target_idx;
+
+	DBG("");
+
+	nla_parse(attr, NFC_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+			genlmsg_attrlen(gnlh, 0), NULL);
+
+	if (attr[NFC_ATTR_DEVICE_INDEX] == NULL)
+		return -ENODEV;
+
+	if (attr[NFC_ATTR_TARGET_INDEX] == NULL)
+		return -ENODEV;
+
+	adapter_idx = nla_get_u32(attr[NFC_ATTR_DEVICE_INDEX]);
+	target_idx = nla_get_u32(attr[NFC_ATTR_TARGET_INDEX]);
+
+	DBG("adapter %d target %d", adapter_idx, target_idx);
+
+	return __near_adapter_remove_target(adapter_idx, target_idx);
+}
+
 static int nfc_netlink_event_dep_up(struct genlmsghdr *gnlh)
 {
 	struct nlattr *attrs[NFC_ATTR_MAX + 1];
@@ -556,7 +561,7 @@ static int nfc_netlink_event_dep_up(struct genlmsghdr *gnlh)
 
 		DBG("%d %d", idx, target_idx);
 
-		return near_adapter_set_dep_state(idx, TRUE);
+		return __near_adapter_set_dep_state(idx, TRUE);
 	} else {
 		return -EOPNOTSUPP;
 	}
@@ -578,11 +583,53 @@ static int nfc_netlink_event_dep_down(struct genlmsghdr *gnlh)
 
 	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
 
-	near_adapter_set_dep_state(idx, FALSE);
+	__near_adapter_set_dep_state(idx, FALSE);
 
 	return 0;
 }
 
+static int nfc_netlink_event_tm_activated(struct genlmsghdr *gnlh)
+{
+	struct nlattr *attrs[NFC_ATTR_MAX + 1];
+	uint32_t idx;
+
+	DBG("");
+
+	nla_parse(attrs, NFC_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+		  genlmsg_attrlen(gnlh, 0), NULL);
+	if (attrs[NFC_ATTR_DEVICE_INDEX] == NULL) {
+		near_error("Missing device index");
+		return -ENODEV;
+	}
+
+	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
+
+	DBG("%d", idx);
+
+	return __near_adapter_add_device(idx, NULL, 0);
+}
+
+static int nfc_netlink_event_tm_deactivated(struct genlmsghdr *gnlh)
+{
+	struct nlattr *attrs[NFC_ATTR_MAX + 1];
+	uint32_t idx;
+
+	DBG("");
+
+	nla_parse(attrs, NFC_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+		  genlmsg_attrlen(gnlh, 0), NULL);
+	if (attrs[NFC_ATTR_DEVICE_INDEX] == NULL) {
+		near_error("Missing device index");
+		return -ENODEV;
+	}
+
+	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
+
+	DBG("%d", idx);
+
+	return __near_adapter_remove_device(idx);
+}
+
 static int nfc_netlink_event(struct nl_msg *n, void *arg)
 {
 	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(n));
@@ -592,7 +639,11 @@ static int nfc_netlink_event(struct nl_msg *n, void *arg)
 	switch (gnlh->cmd) {
 	case NFC_EVENT_TARGETS_FOUND:
 		DBG("Targets found");
-		nfc_netlink_event_targets(gnlh);
+		nfc_netlink_event_targets_found(gnlh);
+		break;
+	case NFC_EVENT_TARGET_LOST:
+		DBG("Target lost");
+		nfc_netlink_event_target_lost(gnlh);
 		break;
 	case NFC_EVENT_DEVICE_ADDED:
 		DBG("Adapter added");
@@ -613,6 +664,16 @@ static int nfc_netlink_event(struct nl_msg *n, void *arg)
 		DBG("DEP link is down");
 		nfc_netlink_event_dep_down(gnlh);
 
+		break;
+	case NFC_EVENT_TM_ACTIVATED:
+		DBG("Target mode activated");
+		nfc_netlink_event_tm_activated(gnlh);
+
+		break;
+	case NFC_EVENT_TM_DEACTIVATED:
+		DBG("Target mode deactivated");
+		nfc_netlink_event_tm_deactivated(gnlh);
+
 		break;
 	}
 
@@ -637,6 +698,8 @@ static gboolean __nfc_netlink_event(GIOChannel *channel,
 
 	nl_recvmsgs(state->nl_sock, cb);
 
+	nl_cb_put(cb);
+
 	return TRUE;
 }
 
@@ -758,17 +821,11 @@ int __near_netlink_init(void)
 		goto handle_destroy;
 	}
 
-	if (genl_ctrl_alloc_cache(nfc_state->nl_sock, &nfc_state->nl_cache)) {
-		near_error("Failed to allocate generic netlink cache");
-		err = -ENOMEM;
-		goto handle_destroy;
-	}
-
-	nfc_state->nlnfc = genl_ctrl_search_by_name(nfc_state->nl_cache, "nfc");
-	if (nfc_state->nlnfc == NULL) {
-		near_error("nfc not found");
+	nfc_state->nfc_id = genl_ctrl_resolve(nfc_state->nl_sock, "nfc");
+	if (nfc_state->nfc_id < 0) {
+		near_error("Unable to find NFC netlink family");
 		err = -ENOENT;
-		goto cache_free;
+		goto handle_destroy;
 	}
 
 	nfc_state->mcid = nl_get_multicast_id(nfc_state->nl_sock, NFC_GENL_NAME,
@@ -776,21 +833,17 @@ int __near_netlink_init(void)
 	if (nfc_state->mcid <= 0) {
 		near_error("Wrong mcast id %d", nfc_state->mcid);
 		err = nfc_state->mcid;
-		goto family_free;
+		goto handle_destroy;
 	}
 
 	err = nl_socket_add_membership(nfc_state->nl_sock, nfc_state->mcid);
 	if (err) {
 		near_error("Error adding nl socket to membership");
-		goto family_free;
+		goto handle_destroy;
 	}
 
 	return nfc_event_listener(nfc_state);
 
-family_free:
-	genl_family_put(nfc_state->nlnfc);
-cache_free:
-	nl_cache_free(nfc_state->nl_cache);
 handle_destroy:
 	nl_socket_free(nfc_state->nl_sock);
 state_free:
@@ -815,8 +868,6 @@ void __near_netlink_cleanup(void)
 	if (nfc_state == NULL)
 		return;
 
-	genl_family_put(nfc_state->nlnfc);
-	nl_cache_free(nfc_state->nl_cache);
 	nl_socket_free(nfc_state->nl_sock);
 
 	g_free(nfc_state);
